//_____________________________________________
//
// KHLAEvent.cxx
// KDataStructure
//
// Author: Adam Cox <mailto:adam.cox@ik.fzk.de>
//
// * Copyright 2010 Karlsruhe Institute of Technology. All rights reserved.
//
// This is the main Event class for High Level Analysis. One typically
// begins all analysis by getting a pointer to this object from 
// the KDataReader object. By calling KDataReader::GetEntry(Int_t i)
// or KDataReader::GetNextEntry, one can loop through the KDS data
// file and examine each event separately. From this HLAEvent object
// one can obtain pointers to they Muon and Bolometer System Records
// GetBoloSystemRecord(void), GetMuonVetoSystemRecord()
// and to each of the event's Sub Records. 

#include "KHLAEvent.h"

#include "TProcessID.h"

//system Record Includes

//sub record includes
#include "TClonesArray.h"
#include "KSambaSubRecord.h"
#include "KSingleBoloSubRecord.h"
#include <iostream>

using namespace std;

ClassImp(KHLAEvent);

KHLAEvent::KHLAEvent(void)
{
	//Default Constructor. This allocates memory for the TClonesArrays
	//and this might be a bug/memory leak. According to the ROOT documentation
	//no memory should be allocated by the default constructor. This needs
	//to be checked. I don't see any memory leak when data processing!!!
	//
	//If you are creating your own KHLAEvent object for use outside 
	//of ROOT I/O, it is strongly encouraged that you use the KEventFactory
	//to make the object for you. It is intended that if the KHLAEvent
	//needs to be built in a more complex way, by always asking the KEventFactory
	//to build an event object for you, you are guaranteed to get a fully
	//created object (assuming that the KEventFactory is properly coded. :)
	
	//HLAEvent members that are on the STACK
	InitializeMembers();
	
	//TClonesArray Initializations must be done separately.
	//as well as any other members that are created on the heap/free store
	fNumSamba = 0;
	fNumBolo = 0;
	
	fSamba = 0;
	fBolo = 0;
	
	fAssignmentOptionNoTref = false;
	
	CreateArrays();
}

KHLAEvent::KHLAEvent(const KHLAEvent &anEvent)
: KEvent(anEvent)
{
	//Copy Constructor. The TRefs will point to newly created objects
	//and not to the data members of the KHLAEvent object you passed in.
	
	CopyLocalMembers(anEvent);
	
	//MUST copy all TRef's appropriately to the new file. Otherwise
	//they will point to records in a different File! This is 
	//why I use the copyClonesArray method here instead of
	//using TClonesArray's copy constructors. 
	fNumSamba = 0;
	fNumBolo = 0;
	
	fSamba = 0;
	fBolo = 0;
	CreateArrays();

	CopyClonesArrays(anEvent);
}

KHLAEvent::~KHLAEvent(void) 
{
	
	Clear("C");
	
	//Delete TClonesArrays
	if(fSamba) delete fSamba;
	fSamba = 0;
	
	if(fBolo) delete fBolo;
	fBolo = 0;
	
	
	
}

KHLAEvent& KHLAEvent::operator=(const KEvent &anEvent)
{
#ifdef _K_DEBUG_EVENT_ASSIGNOP
	cout << "hla = base" << endl;
#endif
	
	//assigns the data member values of an KEvent object into the KHLAEvent. 
	if(&anEvent == this) return *this;

		
	try{
		const KHLAEvent &rhs = dynamic_cast<const KHLAEvent&>(anEvent);
		this->operator=(rhs);
	}
		
	catch(bad_cast) {
		this->KEvent::operator=(anEvent);
	}
	

	return *this;
}



KHLAEvent& KHLAEvent::operator=(const KHLAEvent &anEvent)
{
	//sets anEvent to this. 
	
#ifdef _K_DEBUG_EVENT_ASSIGNOP
	cout << "hla = hla" << endl;
#endif
	if(&anEvent == this) return *this;
	
	this->KEvent::operator=(anEvent);
	
	CopyLocalMembers(anEvent);
	CopyClonesArrays(anEvent);

	return *this;
}

void KHLAEvent::CopyClonesArrays(const KHLAEvent &anEvent)
{
	//
	//ClonesArray assignment doesn't appear to work in the following way
	//*fSamba = *anEvent.GetSambaSubRecords();
	//*fBolo = *anEvent.GetBoloSubRecords();
	//
	//so, I just clear this object's array, and create as many objects as I need.
	//this might be inefficient, but I don't think so. besides,
	//we likely won't be creating many many instances of the Event class.
	//at least, we shouldn't do that. 
	//if your code is really slow and you think it has something
	//to do with copying HLAEvents, look here for possible fixes. 
	//
	//MUST copy all TRef's appropriately to the new file. Otherwise
	//they will point to records in a different File!

	ClearArrays("C");
	
	Int_t ObjectNumber = TProcessID::GetObjectCount();
	
	
	if(fAssignmentOptionNoTref){
		for(Int_t i = 0; i < anEvent.GetNumSambas(); i++){
			KSambaSubRecord *s = AddSamba();
			KSambaSubRecord *sO = anEvent.GetSamba(i);
			if(s != 0 && sO != 0) 
				*s = *sO;
			else
				cout << "KHLAEvent::operator= Invalid Samba Pointer" << endl;
		}
		
		for(Int_t i = 0; i < anEvent.GetNumBolos(); i++){
			KSingleBoloSubRecord *bolo = AddBolo();
			KSingleBoloSubRecord *boloO = anEvent.GetBolo(i);
			
			if(bolo != 0 && boloO != 0) 
				*bolo = *boloO;
			else
				cout << "KHLAEvent::operator= Invalid SingleBolo Pointer" << endl;
		}
		
	}
	else {
	
	
		for(Int_t i = 0; i < anEvent.GetNumBolos(); i++){
			KSingleBoloSubRecord *bolo = AddBolo();
			KSingleBoloSubRecord *boloO = anEvent.GetBolo(i);
			KSambaSubRecord *sambaO = boloO->GetSambaRecord();
			KSambaSubRecord *samba = 0;
			
			if(bolo != 0 && boloO != 0) 
				*bolo = *boloO;
			else
				cout << "KHLAEvent::operator= Invalid SingleBolo Pointer" << endl;
			
			if(sambaO != 0) {
				samba = AddSamba();
				*samba = *sambaO;
				if(samba) bolo->SetSambaRecord(samba);
			}
			else
				cout << "KHLAEvent::operator= No Samba Pointer for bolo record" << endl;
		}
	}
	
	
	//Restore Object count                                                                                                     
	//To save space in the table keeping track of all referenced objects 
	//and computation time,
	//we assume that our events DO NOT address each other. We reset the                                                        
	//object count to what it was at the beginning of the event.                                                               
	TProcessID::SetObjectCount(ObjectNumber);
	
}

void KHLAEvent::CopyLocalMembers(const KHLAEvent &anEvent)
{
	//fDataCleaningWord = anEvent.fDataCleaningWord;
	fNumSamba = anEvent.fNumSamba;
	fNumBolo = anEvent.fNumBolo;
	
}

void KHLAEvent::ClearArrays(Option_t *anOption)
{
	ClearArray(anOption, fSamba,fNumSamba);
	ClearArray(anOption, fBolo,fNumBolo);
}

void KHLAEvent::Clear(Option_t *anOption)
{
	
	ClearArrays(anOption);
	//delete any memory allocated by the KEvent class here that
	//needs to be deleted for each event -- but not the TClonesArrays. Clear should be called 
	//after every event so that the TClonesArrays are cleared. 
	KEvent::Clear(anOption);
	
	//Call the System Record clears in case they need to clean up
	//or to reset initial values of member variables
	
	//reset member variables to zero
	InitializeMembers();
																 
}

void KHLAEvent::InitializeMembers(void)
{
	KEvent::InitializeMembers();
	//fDataCleaningWord.Clear();
	//SetDataType(kHLADataType);
	
}

void KHLAEvent::CreateArrays(void)
{
	//Allocates memory for the TClonesArrays if they haven't already
	//been allocated.
	
	if(!fSamba)
		fSamba = new TClonesArray("KSambaSubRecord",5);
	
	if(!fBolo)
		fBolo = new TClonesArray("KSingleBoloSubRecord",5);
	
	
	//why doesn't this create a memory leak? CreateArrays is called
	//by the default constructor. Does ROOT realize that I already 
	//have these TClonesArrays. Or maybe a TClonesArray object doesn't
	//require that much memory, so I don't notice the leak.
}

void KHLAEvent::ClearArray(Option_t *anOption, TClonesArray *mArray, Int_t &mCount)
{
	if(mArray) {
		mArray->Clear( (anOption && *anOption) ? anOption : "C" );
		//we have to delete because our sub-records contain TString members! :(
		//mArray->Delete( (anOption && *anOption) ? anOption : "C" );
		mCount = 0;
	}
}

template<class T> T* KHLAEvent::AddSubRecord(TClonesArray *mArray, Int_t &mCount)
{
	TClonesArray &mArrayRef = *mArray;
	T* mNewSubRecord = new(mArrayRef[mCount++]) T;
	return mNewSubRecord;
}

KSambaSubRecord* KHLAEvent::AddSamba(void)
{
	//Use this event only when creating an event and you want to add
	//a new SubRecord.
	
	return AddSubRecord<KSambaSubRecord>(fSamba,fNumSamba);
}

KSingleBoloSubRecord* KHLAEvent::AddBolo(void)
{
	//Use this event only when creating an event and you want to add
	//a new SubRecord.
	
	AddTriggerType(kBoloTriggerType);
	return AddSubRecord<KSingleBoloSubRecord>(fBolo,fNumBolo);
}


KSambaSubRecord *KHLAEvent::GetSamba(Int_t i) const
{
  // Return the i'th Samba Sub Record for this event.
	
  KSambaSubRecord *ms = 0;
  if (i < fNumSamba) ms = (KSambaSubRecord *)fSamba->At(i);
  return ms;
}

KSingleBoloSubRecord *KHLAEvent::GetBolo(Int_t i) const
{
  // Return the i'th Bolometer Sub Record for this event.
	
  KSingleBoloSubRecord *ms = 0;
  if (i < fNumBolo) ms = (KSingleBoloSubRecord *)fBolo->At(i);
  return ms;
}


Bool_t KHLAEvent::IsSame(const KHLAEvent &anEvent, Bool_t bPrint) const
{
	//Compares two KHLAEvents and their member variables to test for equality.
	//If bPrint is set to true, then a message for each member variable that is different
	//will print to standard out. Otherwise, this method will return false and quit
	//checking member variables as soon as it finds a unequal data member.
	
	Bool_t bIsEqual = true; //assume its true, then test for differences
	
	//call the base class's IsSame methods
	if(!this->KEvent::IsSame(anEvent,bPrint)){
		if (bPrint) 
		bIsEqual = false;
		if(!bPrint)
			return false;  //if we're not printing out, just return false at first failure
		//the operator== method uses this functionality.
	}
	
	/*if(fDataCleaningWord != anEvent.fDataCleaningWord){
		if (bPrint) 
			cout << "KHLAEvent fDataCleaningWord Not Equal" << endl;		
		bIsEqual = false;
		if(!bPrint)
			return false;  //if we're not printing out, just return false at first failure
		//the operator== method uses this functionality.
	}*/
	
	
	
	if(fNumSamba == anEvent.fNumSamba){
		for(Int_t i = 0; i < fNumSamba; i++){
			KSambaSubRecord *s = GetSamba(i);
			KSambaSubRecord *sOther = anEvent.GetSamba(i);
			if(s != 0 && sOther != 0){
				if(!s->IsSame(*sOther, bPrint)){
					if (bPrint) 
						cout << "KHLAEvent KSambaSubRecord number " << i << " Not Equal" << endl;		
					bIsEqual = false;
					if(!bPrint)
						return false;
				}
			}
		}
	}
	else {
		if (bPrint) 
			cout << "KHLAEvent fNumSamba Not Equal" << endl;		
		bIsEqual = false;
		if(!bPrint)
			return false;
	}
	
	if(fNumBolo == anEvent.fNumBolo){
		for(Int_t i = 0; i < fNumBolo; i++){
			KSingleBoloSubRecord *s = GetBolo(i);
			KSingleBoloSubRecord *sOther = anEvent.GetBolo(i);
			if(s != 0 && sOther != 0){
				if(!s->IsSame(*sOther, bPrint)){
					if (bPrint) 
						cout << "KHLAEvent KSingleBoloSubRecord number " << i << " Not Equal" << endl;		
					bIsEqual = false;
					if(!bPrint)
						return false;
				}
			}
		}
	}
	else {
		if (bPrint) 
			cout << "KHLAEvent fNumBolo Not Equal" << endl;		
		bIsEqual = false;
		if(!bPrint)
			return false;
	}
	
	
		
	return bIsEqual;
}



void KHLAEvent::Compact(void)
{
	//Make the event class as small as possible. This calls 'Compact' for all member
	//variables that are KDS classes, member variables that can be compacted (such as TBits)
	//and base classes
	
	KEvent::Compact();
	
	
	for(Int_t i = 0; i < GetNumSambas(); i++){
		KSambaSubRecord* samba = GetSamba(i);
		samba->Compact();
	}
	for(Int_t i = 0; i < GetNumBolos(); i++){
		KSingleBoloSubRecord* bolo = GetBolo(i);
		bolo->Compact();
	}

}
